#include "Play.hpp"
#include <utility>

#include <Vector/Vector.hpp>

using std::swap;

void Block::create_body(dWorldID world, dSpaceID space) {
	assert(body == NULL);
	assert(geom == NULL);

	body = dBodyCreate(world);
	{ //mass for body:
		dMass mass;
		dMassSetBox(&mass, 1.0f, radius.x*2, radius.y*2, radius.z*2);
		dBodySetMass(body, &mass);
	}
	//geom for body:
	geom = dCreateBox(space, radius.x*2, radius.y*2, radius.z*2);
	dGeomSetBody(geom, body);
	dBodySetPosition(body, position.x, position.y, position.z);
	dBodySetData(body, this);
	{ //copy the orientation:
		dQuaternion quat;
		quat[1] = orientation[0];
		quat[2] = orientation[1];
		quat[3] = orientation[2];
		quat[0] = orientation[3];
		dBodySetQuaternion(body, quat);
	}
}

void Block::destroy_body() {
	assert(body);
	assert(geom);
	dGeomDestroy(geom);
	geom = NULL;
	dBodyDestroy(body);
	body = NULL;
}

void Block::copy_body_position() {
	const dReal *pos = dBodyGetPosition(body);
	position.x = pos[0];
	position.y = pos[1];
	position.z = pos[2];
	const dReal *quat = dBodyGetQuaternion(body);
	orientation[0] = quat[1];
	orientation[1] = quat[2];
	orientation[2] = quat[3];
	orientation[3] = quat[0];
}

void Block::reset_body_position() {
	dBodySetPosition(this->body, position.x, position.y, position.z);
}

void Play::start_simulation() {
	assert(world == NULL);
	assert(space == NULL);
	assert(collisions == NULL);
	world = dWorldCreate();

	dBodyCreate(world);


	space = dHashSpaceCreate(NULL); //param is space to add it to.
	ground = dCreatePlane(space, 0.0f, 0.0f, 1.0f, 0.0f); //plane with z = 0
	collisions = dJointGroupCreate(0); //param is unused and should be zero


	for (unsigned int b = 0; b < world_blocks.size(); ++b) {
		world_blocks[b].create_body(world, space);
		world_blocks[b].in_stack = false;
	}

	for (unsigned int b = 0; b < black_holes.size(); ++b) {
		black_holes[b].create_body(world, space);
		black_holes[b].in_stack = false;
	}

	dWorldSetGravity(world, 0.0f, 0.0f, -10.0f);

	sim_acc = 0.0f;
}

void Play::stop_simulation() {
	assert(world != NULL);
	assert(space != NULL);
	assert(collisions != NULL);

	for (unsigned int b = 0; b < world_blocks.size(); ++b) {
		world_blocks[b].destroy_body();
	}

	for (unsigned int b = 0; b < black_holes.size(); ++b) {
		black_holes[b].destroy_body();
	}

	dGeomDestroy(ground);
	ground = NULL;

	dJointGroupDestroy(collisions);
	collisions = NULL;

	dSpaceDestroy(space);
	space = NULL;

	dWorldDestroy(world);
	world = NULL;
}

class Info {
public:
	dWorldID world;
	dJointGroupID contacts;
	vector< Block > *blocks;
};

//abducted from ode boxstack demo:
void near_callback (void *data, dGeomID o1, dGeomID o2) {
	Info *info = (Info *)data;

	const unsigned int MAX_CONTACTS = 10;

	dContact contact[MAX_CONTACTS];   // up to MAX_CONTACTS contacts per box-box
	for (unsigned int i=0; i<MAX_CONTACTS; ++i) {
		contact[i].surface.mode = dContactBounce | dContactSoftCFM;
		contact[i].surface.mu = dInfinity;
		contact[i].surface.mu2 = 0;
		contact[i].surface.bounce = 0.1;
		contact[i].surface.bounce_vel = 0.1;
		contact[i].surface.soft_cfm = 0.01;
	}
	dBodyID b1 = dGeomGetBody(o1);
	dBodyID b2 = dGeomGetBody(o2);
	if (int numc = dCollide(o1, o2, MAX_CONTACTS, &contact[0].geom, sizeof(dContact))) {
		for (int i=0; i<numc; i++) {
			dJointID c = dJointCreateContact(info->world,info->contacts,contact+i);
			dJointAttach (c,b1,b2);
		}
	}

	//tried to delete any blocks that hit the black holes, but this method doesn't have access to the vectors I want

	/*for (unsigned int a = 0; a < black_holes.size(); ++a) {
		for (unsigned int b = 0; b < world_blocks.size(); ++b) {
			if (int numc = dCollide(o1, o2, MAX_CONTACTS, &contact[0].geom, sizeof(dContact))) {
				if(b1 == world_blocks[b].body && b2 == black_holes[a].body) {
					world_blocks.remove(b);
				}else if(b2 == world_blocks[b].body && b1 == black_holes[a].body) {
					world_blocks.remove(b);
				}
			}
		}
	}*/
}

void Play::tick_simulation() {

	Info info;
	info.world = world;
	info.contacts = collisions;
	info.blocks = &world_blocks;

	for (unsigned int a = 0; a < black_holes.size(); ++a) {
		for (unsigned int b = 0; b < world_blocks.size(); ++b) {
			Vector3f force = black_holes[a].position - world_blocks[b].position;
			float len = length(force);
			force = 3.0f*force/(len*len);

			dBodyAddForce(world_blocks[b].body, force.x, force.y, force.z);
		}
	}

	dSpaceCollide(space, &info, near_callback);

	dWorldQuickStep(world, Tick);
	dJointGroupEmpty(collisions);

	for (unsigned int b = 0; b < world_blocks.size(); ++b) {
		world_blocks[b].copy_body_position();
	}
	for (unsigned int b = 0; b < black_holes.size(); ++b) {
		black_holes[b].reset_body_position();
	}
}
