#include "Play.hpp"

#include <Graphics/Graphics.hpp>
#include <Graphics/ShadowStencil.hpp>
#include <Vector/VectorGL.hpp>
#include <Vector/QuatGL.hpp>
#include <Vector/Box.hpp>

#include <Graphics/GLSLShader.hpp>

#include <tr1/unordered_map>
#include <utility>

using std::tr1::unordered_map;
using std::pair;
using std::make_pair;

using namespace std;

//hash function specialization for pair< >,
// actually kind of expected this to be in STL, but isn't with the version
// of gcc I'm currently using.
namespace std {
namespace tr1 {
template< typename A, typename B >
class hash< std::pair< A, B > > {
public:
	size_t operator()(const std::pair< A, B > &x) const {
		size_t ha, hb;
		{
			std::tr1::hash< A > HA;
			ha = HA(x.first);
		}
		{
			std::tr1::hash< B > HB;
			hb = HB(x.second);
		}
		return ((ha << (sizeof(size_t)/2)) | (ha >> (sizeof(size_t) - sizeof(size_t)/2))) ^ hb;
	}
};
}
}

REQUIRE_GL_EXTENSION(GL_ARB_shader_objects);

Graphics::ProgramObjectRef normal_displace_shader = NULL;
Graphics::ProgramObjectRef no_displace_shader = NULL;

void load_shaders() {
	if (normal_displace_shader.ref) return;
	{
		Graphics::ShaderObjectRef vert = Graphics::get_shader_object("shaders/normal_displace.glsl", GL_VERTEX_SHADER_ARB);
		assert(vert.ref);
		normal_displace_shader = Graphics::get_program_object(vert);
		assert(normal_displace_shader.ref);
	}
	{
		Graphics::ShaderObjectRef vert = Graphics::get_shader_object("shaders/no_displace.glsl", GL_VERTEX_SHADER_ARB);
		assert(vert.ref);
		no_displace_shader = Graphics::get_program_object(vert);
		assert(no_displace_shader.ref);
	}
}

void Block::draw() {

	glPushMatrix();
	glTranslatef(position.x, position.y, position.z);
	glRotate(orientation);

	Vector3f verts[8];
	for (unsigned int i = 0; i < 8; ++i) {
		verts[i] = -1.0f * radius;
		if (i & 1) verts[i].x *= -1.0f;
		if (i & 2) verts[i].y *= -1.0f;
		if (i & 4) verts[i].z *= -1.0f;
	}
	Vector4ui faces[6] = {
		make_vector( 0U, 4U, 6U, 2U ),
		make_vector( 0U, 2U, 3U, 1U ),
		make_vector( 0U, 1U, 5U, 4U ),
		make_vector( 7U, 6U, 4U, 5U ),
		make_vector( 7U, 3U, 2U, 6U ),
		make_vector( 7U, 5U, 1U, 3U ),
	};

	vector< Vector3f > draw_verts;
	vector< Vector3f > draw_norms;

	unordered_map< pair< unsigned int, unsigned int >, Vector3f > edge_norms;

	for (unsigned int f = 0; f < 6; ++f) {
		Vector3f norm = normalize(cross_product(
			verts[faces[f][1]] - verts[faces[f][0]],
			verts[faces[f][3]] - verts[faces[f][0]]));
		for (unsigned int i = 0; i < 4; ++i) {
			draw_norms.push_back(norm);
			draw_verts.push_back(verts[faces[f][i]]);
			pair< unsigned int, unsigned int > edge = make_pair(faces[f][i], faces[f][(i+1)%4]);
			assert(!edge_norms.count(edge));
			edge_norms.insert(make_pair(edge, norm));
		}
	}
	for (typeof(edge_norms.begin()) e = edge_norms.begin(); e != edge_norms.end(); ++e) {
		if (e->first.first > e->first.second) continue;
		pair< unsigned int, unsigned int > rev = make_pair(e->first.second, e->first.first);
		if (!edge_norms.count(rev)) continue;
		draw_norms.push_back(e->second);
		draw_verts.push_back(verts[e->first.second]);
		draw_norms.push_back(e->second);
		draw_verts.push_back(verts[e->first.first]);

		draw_norms.push_back(edge_norms[rev]);
		draw_verts.push_back(verts[e->first.first]);
		draw_norms.push_back(edge_norms[rev]);
		draw_verts.push_back(verts[e->first.second]);
	}

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);

	glVertexPointer(3, GL_FLOAT, sizeof(Vector3f), &draw_verts[0]);
	glNormalPointer(GL_FLOAT, sizeof(Vector3f), &draw_norms[0]);

	assert(draw_verts.size() == draw_norms.size());
	glDrawArrays(GL_QUADS, 0, draw_verts.size());

	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);

	glPopMatrix();

}

Box3f Block::get_bounds() {
	Box3f bounds = make_box(make_vector(0.0f, 0.0f, 0.0f), make_vector(0.0f, 0.0f, 0.0f));
	for (unsigned int i = 0; i < 8; ++i) {
		Vector3f p = radius;
		if (i & 1) p.x *= -1.0f;
		if (i & 2) p.y *= -1.0f;
		if (i & 4) p.z *= -1.0f;
		p = rotate(orientation, p);
		bounds.min = min(bounds.min, p);
		bounds.max = max(bounds.max, p);
	}
	bounds.min += position;
	bounds.max += position;
	return bounds;
}


Play::Play() {
	camera_at = make_vector(5.0f, 5.0f, 10.0f);
	camera_target = make_vector(0.0f, 0.0f, 0.0f);

	camera_move = make_vector(0.0f, 0.0f);
	block_move = make_vector(0.0f, 0.0f);

	generate_level();
	start_stacking();
	world = NULL;
	space = NULL;
	ground = NULL;
	collisions = NULL;
}

Play::~Play() {
}

void Play::generate_level() {
	for (int x = -4; x <= 4; ++x) {
		for (int y = -4; y <= 4; ++y) {
			if (x == 0 && y == 0) continue;
			world_blocks.push_back(Block());
			world_blocks.back().radius = make_vector(0.1f + rand() / float(RAND_MAX) * 0.2f, 0.1f + rand() / float(RAND_MAX) * 0.2f, 0.1f + rand() / float(RAND_MAX) * 0.4f);
			world_blocks.back().orientation = rotation(rand() / float(RAND_MAX) * float(M_PI) * 2.0f, make_vector(0.0f, 0.0f, 1.0f));
			world_blocks.back().position = make_vector((float)x, (float)y, 0.0f + world_blocks.back().radius.z);
		}
	}
}

void Play::start_stacking() {
	stacking = true;
}

void infinite_perspective(float fovy, float aspect, float near) {
	//with help from:
	//http://steinsoft.net/index.php?site=Programming/Code%20Snippets/OpenGL/gluperspective
	float left, right, top, bottom;
	top = near * tan(fovy / 360.0f * ((float)M_PI));
	bottom = -top;
	left = bottom * aspect;
	right = top * aspect;

	//with help from nvidia's robust stencil shadows presentation also:
	float mat[16] = {
		2.0f * near / (right - left), 0, 0, 0,
		0, 2.0f * near / (top - bottom), 0, 0,
		(right + left) / (right - left), (top + bottom) / (top - bottom), -1.0f, -1.0f,
		0, 0, -2.0f * near, 0
	};
	glMultMatrixf(mat);

}

void Play::draw() {
	glClearColor(0.8f, 0.8f, 0.8f, 0.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_LINE_SMOOTH);
	glLineWidth(2.0f);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	infinite_perspective(90.0f, Graphics::aspectf(), 0.2f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt(camera_at.x, camera_at.y, camera_at.z, camera_target.x, camera_target.y, camera_target.z, 0.0f, 0.0f, 1.0f);

	for (unsigned int pass = 0; pass <= 2; ++pass) {

		static Vector4f l = normalize(make_vector(1.0f, 3.0f, 10.0f, 0.0f));
		static Vector4f amb = make_vector(0.1f, 0.1f, 0.1f, 1.0f);
		static Vector4f dif = make_vector(1.0f, 1.0f, 1.0f, 1.0f);
		static Vector4f smalldif = make_vector(0.2f, 0.2f, 0.2f, 1.0f);
		static Vector4f zero = make_vector(0.0f, 0.0f, 0.0f, 0.0f);
		static Vector4f one = make_vector(1.0f, 1.0f, 1.0f, 1.0f);
		if (pass == 0) {
			//Ambient only:
			glEnable(GL_LIGHTING);
			glEnable(GL_LIGHT0);
			glDisable(GL_LIGHT0);
			glLightfv(GL_LIGHT0, GL_AMBIENT, amb.c);
			glLightfv(GL_LIGHT0, GL_DIFFUSE, smalldif.c);
			glLightfv(GL_LIGHT0, GL_POSITION, l.c);
			glEnable(GL_CULL_FACE);
			glCullFace(GL_BACK);
		} else if (pass == 1) {
			//Shadow volumes:
			glDisable(GL_LIGHTING);
			glDisable(GL_CULL_FACE);
			glEnable(GL_STENCIL_TEST);
			Graphics::enable_shadow_stencil();
			glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
			glDepthMask(GL_FALSE);
			glDepthFunc(GL_LESS);
		} else if (pass == 2) {
			//Lit faces:
			Graphics::disable_shadow_stencil();
			glStencilFunc(GL_EQUAL, 0, ~0U);
			glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
			glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
			glDepthMask(GL_TRUE);

			glEnable(GL_LIGHTING);
			glEnable(GL_LIGHT0);
			glLightfv(GL_LIGHT0, GL_AMBIENT, amb.c);
			glLightfv(GL_LIGHT0, GL_DIFFUSE, dif.c);
			glLightfv(GL_LIGHT0, GL_POSITION, l.c);
			glEnable(GL_CULL_FACE);
			glCullFace(GL_BACK);
			glDepthFunc(GL_LEQUAL);
		}

		load_shaders();
		if (pass == 1) {
			glUseProgramObjectARB(normal_displace_shader->handle);
		} else {
			glUseProgramObjectARB(no_displace_shader->handle);
		}

		if (pass != 1) {
			if (pass == 2) {
				glLightfv(GL_LIGHT0, GL_DIFFUSE, zero.c);
				glLightfv(GL_LIGHT0, GL_AMBIENT, one.c);
			}
			glColor3f(0.8f, 0.8f, 0.8f);
			glNormal3f(0.0f, 0.0f, 1.0f);
			glBegin(GL_QUADS);
			glVertex2f(-10.0f, -10.0f);
			glVertex2f( 10.0f, -10.0f);
			glVertex2f( 10.0f,  10.0f);
			glVertex2f(-10.0f,  10.0f);
			glEnd();
			glColor3f(1.0f, 1.0f, 1.0f);
			glBegin(GL_LINES);
			for (int x = -5; x <= 4; ++x) {
				glVertex3f(x+0.5f, -4.5f, 0.01f);
				glVertex3f(x+0.5f,  4.5f, 0.01f);
				glVertex3f(-4.5f, x+0.5f, 0.01f);
				glVertex3f( 4.5f, x+0.5f, 0.01f);
			}
			glEnd();
			if (pass == 2) {
				glLightfv(GL_LIGHT0, GL_DIFFUSE, dif.c);
				glLightfv(GL_LIGHT0, GL_AMBIENT, amb.c);
			}
		}

		
		glColor3f(1.0f, 1.0f, 1.0f);
		for (unsigned int b = 0; b < world_blocks.size(); ++b) {
			world_blocks[b].draw();
		}

		glColor3f(0.0f, 0.0f, 0.0f);
		for (unsigned int b = 0; b < black_holes.size(); ++b) {
			black_holes[b].draw();
		}

		glUseProgramObjectARB(0);
	}
	glDisable(GL_LIGHTING);
	glDisable(GL_STENCIL_TEST);
	glDepthFunc(GL_LESS);


	Graphics::gl_errors("draw");
}

void Play::update(float elapsed_time) {
	{ //update camera.
		Vector3f out = camera_at - camera_target;
		float radius = length(out);
		out = normalize(out);
		Vector3f up = make_vector(0.0f, 0.0f, 1.0f);
		up = normalize(up - out * (up * out));
		Vector3f right = cross_product(out, up);
		out = normalize(out + right * elapsed_time * camera_move.x + up * elapsed_time * camera_move.y);
		if (out.z > 0.9f) {
			out.z = 0.9f;
			out.xy = normalize(out.xy) * sqrtf(1.0f - 0.9f * 0.9f);
		}
		if (out.z < 0.1f) {
			out.z = 0.1f;
			out.xy = normalize(out.xy) * sqrtf(1.0f - 0.1f * 0.1f);
		}
		camera_at = camera_target + out * radius;
	}

	if (world) {
		sim_acc += elapsed_time;
		while (sim_acc > 0.0f) {
			sim_acc -= Tick;
			tick_simulation();
		}
	}

}

void do_move(Vector4i dirs, SDL_Event const &event, Vector2f &move) {
	float amt = 1.0f;
	if (event.type == SDL_KEYDOWN) {
		amt = 1.0f;
	} else if (event.type == SDL_KEYUP) {
		amt =-1.0f;
	} else {
		return;
	}
	if (event.key.keysym.sym == dirs[0]) move.y += amt;
	if (event.key.keysym.sym == dirs[1]) move.y -= amt;
	if (event.key.keysym.sym == dirs[2]) move.x -= amt;
	if (event.key.keysym.sym == dirs[3]) move.x += amt;
	move = max(make_vector(-1.0f,-1.0f), min(make_vector(1.0f, 1.0f), move));
}

void Play::handle_event(SDL_Event const &event) {
	//Handle events here:
	do_move(make_vector< int >(SDLK_UP, SDLK_DOWN, SDLK_LEFT, SDLK_RIGHT), event, camera_move);
	do_move(make_vector< int >(SDLK_w, SDLK_s, SDLK_a, SDLK_d), event, block_move);
	if (event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_RETURN) {
		if (stacking) {
			stacking = false;
			old_world_blocks = world_blocks;
			start_simulation();
		} else {
			stop_simulation();
			start_stacking();
		}
	}

	if (event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_SPACE) {
		if (!stacking) {
			stop_simulation();
			start_stacking();
		}

		black_holes.push_back(Block());
		black_holes.back().radius = make_vector(0.2f, 0.2f, 0.2f);
		black_holes.back().orientation = rotation(0.0f, make_vector(0.0f, 0.0f, 0.0f));
		black_holes.back().position = make_vector(-5.0f, -5.0f, 0.0f);
	}

	if (event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_a) {
		if (!stacking) {
			stop_simulation();
			start_stacking();
		}

		if(black_holes.size() > 0)
			black_holes.back().position.x += 1.0f;
	}
	if (event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_d) {
		if (!stacking) {
			stop_simulation();
			start_stacking();
		}

		if(black_holes.size() > 0)
			black_holes.back().position.x -= 1.0f;
	}
	if (event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_w) {
		if (!stacking) {
			stop_simulation();
			start_stacking();
		}

		if(black_holes.size() > 0)
			black_holes.back().position.y -= 1.0f;
	}
	if (event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_s) {
		if (!stacking) {
			stop_simulation();
			start_stacking();
		}

		if(black_holes.size() > 0)
			black_holes.back().position.y += 1.0f;
	}

	if (event.type == SDL_KEYDOWN) {
		if (event.key.keysym.sym == SDLK_ESCAPE) {
			//escape quits.
			quit_screen(this);
			return;
		}
	}
}


